@@ -1,6 +1,6 @@
 /*
  *  Phusion Passenger - https://www.phusionpassenger.com/
- *  Copyright (c) 2011-2015 Phusion Holding B.V.
+ *  Copyright (c) 2011-2016 Phusion Holding B.V.
  *
  *  "Passenger", "Phusion Passenger" and "Union Station" are registered
  *  trademarks of Phusion Holding B.V.
@@ -295,6 +295,7 @@ Controller::onAppSourceData(Client *client, Request *req, const MemoryKit::mbuf
 			// EOF
 			UPDATE_TRACE_POINT();
 			SKC_TRACE(client, 2, "Application sent EOF");
+			SKC_TRACE(client, 2, "Not keep-aliving application session connection");
 			req->session->close(true, false);
 			endRequest(&client, &req);
 			return Channel::Result(0, false);
@@ -331,15 +332,15 @@ Controller::onAppResponseBegin(Client *client, Request *req) {
 	resp->date = resp->headers.lookup(HTTP_DATE);
 	resp->setCookie = resp->headers.lookup(ServerKit::HTTP_SET_COOKIE);
 	if (resp->setCookie != NULL) {
-		// Remove Set-Cookie from resp->headers without deallocating it.
+		// Move the Set-Cookie header from resp->headers to resp->setCookie;
+		// remove Set-Cookie from resp->headers without deallocating it.
 		LString *copy;
 
 		copy = (LString *) psg_palloc(req->pool, sizeof(LString));
-		*copy = *resp->setCookie;
+		psg_lstr_init(copy);
+		psg_lstr_move_and_append(resp->setCookie, req->pool, copy);
 
-		resp->setCookie->start = NULL;
-		resp->setCookie->end = NULL;
-		resp->setCookie->size = 0;
+		P_ASSERT_EQ(resp->setCookie->size, 0);
 		psg_lstr_append(resp->setCookie, req->pool, "x", 1);
 		resp->headers.erase(ServerKit::HTTP_SET_COOKIE);
 
@@ -821,6 +822,7 @@ Controller::sendResponseHeaderWithWritev(Client *client, Request *req,
 	if (constructHeaderBuffersForResponse(req, buffers,
 		maxbuffers, nbuffers, dataSize, nCacheableBuffers))
 	{
+		UPDATE_TRACE_POINT();
 		SKC_TRACE(client, 2, "Sending response headers using writev()");
 		logResponseHeaders(client, req, buffers, nbuffers, dataSize);
 		markHeaderBuffersForTurboCaching(client, req, buffers, nCacheableBuffers);
@@ -833,6 +835,7 @@ Controller::sendResponseHeaderWithWritev(Client *client, Request *req,
 		req->responseBegun |= ret > 0;
 		return ret == (ssize_t) dataSize;
 	} else {
+		UPDATE_TRACE_POINT();
 		bytesWritten = 0;
 		return false;
 	}
@@ -842,6 +845,7 @@ void
 Controller::sendResponseHeaderWithBuffering(Client *client, Request *req,
 	unsigned int offset)
 {
+	TRACE_POINT();
 	struct iovec *buffers;
 	unsigned int nbuffers, dataSize, nCacheableBuffers;
 	bool ok;
@@ -857,18 +861,21 @@ Controller::sendResponseHeaderWithBuffering(Client *client, Request *req,
 	assert(ok);
 	(void) ok; // Shut up compiler warning
 
+	UPDATE_TRACE_POINT();
 	logResponseHeaders(client, req, buffers, nbuffers, dataSize);
 	markHeaderBuffersForTurboCaching(client, req, buffers, nCacheableBuffers);
 
 	MemoryKit::mbuf_pool &mbuf_pool = getContext()->mbuf_pool;
 	const unsigned int MBUF_MAX_SIZE = mbuf_pool_data_size(&mbuf_pool);
 	if (dataSize <= MBUF_MAX_SIZE) {
+		UPDATE_TRACE_POINT();
 		SKC_TRACE(client, 2, "Sending response headers using an mbuf");
 		MemoryKit::mbuf buffer(MemoryKit::mbuf_get(&mbuf_pool));
 		gatherBuffers(buffer.start, MBUF_MAX_SIZE, buffers, nbuffers);
 		buffer = MemoryKit::mbuf(buffer, offset, dataSize - offset);
 		writeResponse(client, buffer);
 	} else {
+		UPDATE_TRACE_POINT();
 		SKC_TRACE(client, 2, "Sending response headers using a psg_pool buffer");
 		char *buffer = (char *) psg_pnalloc(req->pool, dataSize);
 		gatherBuffers(buffer, dataSize, buffers, nbuffers);
@@ -881,6 +888,7 @@ Controller::logResponseHeaders(Client *client, Request *req, struct iovec *buffe
 	unsigned int nbuffers, unsigned int dataSize)
 {
 	if (OXT_UNLIKELY(getLogLevel() >= LVL_DEBUG3)) {
+		TRACE_POINT();
 		char *buffer = (char *) psg_pnalloc(req->pool, dataSize);
 		gatherBuffers(buffer, dataSize, buffers, nbuffers);
 		SKC_TRACE(client, 3, "Sending response headers: \"" <<
@@ -888,6 +896,7 @@ Controller::logResponseHeaders(Client *client, Request *req, struct iovec *buffe
 	}
 
 	if (req->useUnionStation()) {
+		TRACE_POINT();
 		const char *status = getStatusCodeAndReasonPhrase(req->appResponse.statusCode);
 		if (status != NULL) {
 			req->logMessage("Status: " + StaticString(status));
@@ -898,6 +907,7 @@ Controller::logResponseHeaders(Client *client, Request *req, struct iovec *buffe
 		if (req->appResponse.statusCode >= 400 && req->appResponse.statusCode <= 599) {
 			// Log the request headers like Request headers: { header1: values1-concatenated, header2: values2-concatenated } (single line)
 			// Concatenation was already done by HeaderTable.h:insert (using a comma ',' for joining, or a semicolon ';' for Cookie headers
+			UPDATE_TRACE_POINT();
 			ServerKit::HeaderTable::Iterator it(req->headers);
 			Json::Value json;
 			while (*it != NULL) {
@@ -1078,7 +1088,24 @@ Controller::handleAppResponseBodyEnd(Client *client, Request *req) {
 
 OXT_FORCE_INLINE void
 Controller::keepAliveAppConnection(Client *client, Request *req) {
-	req->session->close(true, req->appResponse.wantKeepAlive);
+	if (req->halfClosePolicy == Request::HALF_CLOSE_PERFORMED) {
+		SKC_TRACE(client, 2, "Not keep-aliving application session connection"
+			" because it had been half-closed before");
+		req->session->close(true, false);
+	} else {
+		// halfClosePolicy is initialized in sendHeaderToApp(). That method is
+		// called immediately after checking out a session, before any events
+		// from the appSource channel can be received.
+		assert(req->halfClosePolicy != Request::HALF_CLOSE_POLICY_UNINITIALIZED);
+		if (req->appResponse.wantKeepAlive) {
+			SKC_TRACE(client, 2, "Keep-aliving application session connection");
+			req->session->close(true, true);
+		} else {
+			SKC_TRACE(client, 2, "Not keep-aliving application session connection"
+				" because application did not allow it");
+			req->session->close(true, false);
+		}
+	}
 }
 
 void
